-- This module provides the definitions for the gates used in our
-- various sets of generators for Clifford+CS operators, and some
-- convenient abbreviations.

open import Lib.Word

module Gate where

data Gate : Set where

  -- Permutation generators.
  CCX0-gen : Gate
  CCX1-gen : Gate
  CCX2-gen : Gate
  CX01-gen : Gate
  CX10-gen : Gate
  CX12-gen : Gate
  CX21-gen : Gate
  CX02-gen : Gate
  CX20-gen : Gate
  X0-gen : Gate
  X1-gen : Gate
  X2-gen : Gate
  Swap01-gen : Gate
  Swap12-gen : Gate

  -- Diagonal generators.
  S0-gen : Gate
  S1-gen : Gate
  S2-gen : Gate
  CS01-gen : Gate
  CS12-gen : Gate
  CS02-gen : Gate
  CCZ-gen : Gate
  iI-gen : Gate

  -- K, CK, CCK' generators.
  CCK'-gen : Gate
  CK10-gen : Gate
  CK20-gen : Gate
  K0-gen : Gate

  -- Not used in Step0, Step3, (Step4?)
  K1-gen : Gate
  K2-gen : Gate

-- Abbreviations.
CCX0 : Word Gate
CCX0 = gen CCX0-gen

CCX1 : Word Gate
CCX1 = gen CCX1-gen

CCX2 : Word Gate
CCX2 = gen CCX2-gen

CX01 : Word Gate
CX01 = gen CX01-gen

CX10 : Word Gate
CX10 = gen CX10-gen

CX12 : Word Gate
CX12 = gen CX12-gen

CX21 : Word Gate
CX21 = gen CX21-gen

CX02 : Word Gate
CX02 = gen CX02-gen

CX20 : Word Gate
CX20 = gen CX20-gen

X0 : Word Gate
X0 = gen X0-gen

X1 : Word Gate
X1 = gen X1-gen

X2 : Word Gate
X2 = gen X2-gen

Swap01 : Word Gate
Swap01 = gen Swap01-gen

Swap12 : Word Gate
Swap12 = gen Swap12-gen

S0 : Word Gate
S0 = gen S0-gen

S1 : Word Gate
S1 = gen S1-gen

S2 : Word Gate
S2 = gen S2-gen

CS01 : Word Gate
CS01 = gen CS01-gen

CS12 : Word Gate
CS12 = gen CS12-gen

CS02 : Word Gate
CS02 = gen CS02-gen

CCZ : Word Gate
CCZ = gen CCZ-gen

iI : Word Gate
iI = gen iI-gen

CCK' : Word Gate
CCK' = gen CCK'-gen

CK10 : Word Gate
CK10 = gen CK10-gen

CK20 : Word Gate
CK20 = gen CK20-gen

K0 : Word Gate
K0 = gen K0-gen

K1 : Word Gate
K1 = gen K1-gen

K2 : Word Gate
K2 = gen K2-gen

