open import Lib.Equality as Eq using (_==_)
open import Lib.Maybe
open import Lib.Word
open import Lib.Nat
open import Lib.MaybeEq
open import Lib.MaybeEq-Instances

open import Index

open import Step1.Theorem using (module TwoLevel-Less)

module Step2.Theorem where

module TwoLevel-Simplified where
  -- The new set of generators.
  data Gen : Set where
    i₀-gen : Gen
    K₀₁-gen : Gen
    X₀₁-gen : Gen
    X₁₂-gen : Gen
    X₂₃-gen : Gen
    X₃₄-gen : Gen
    X₄₅-gen : Gen
    X₅₆-gen : Gen
    X₆₇-gen : Gen

  instance
    MaybeEq-Gen : MaybeEq Gen
    MaybeEq-Gen = record { _=m?_ = λ{
      i₀-gen i₀-gen → Just Eq.refl ;
      K₀₁-gen K₀₁-gen → Just Eq.refl ;
      X₀₁-gen X₀₁-gen → Just Eq.refl ;
      X₁₂-gen X₁₂-gen → Just Eq.refl ;
      X₂₃-gen X₂₃-gen → Just Eq.refl ;
      X₃₄-gen X₃₄-gen → Just Eq.refl ;
      X₄₅-gen X₄₅-gen → Just Eq.refl ;
      X₅₆-gen X₅₆-gen → Just Eq.refl ;
      X₆₇-gen X₆₇-gen → Just Eq.refl ;
      x y -> Nothing } }
  

  -- For convenience, we define a singleton word for each new
  -- generator, as well as for all of the original 2-level generators.
  -- This makes the relations easier to state.

  mX-aux : Index -> Index -> Word Gen
  mX-aux ₀ ₁ = gen X₀₁-gen
  mX-aux ₁ ₂ = gen X₁₂-gen
  mX-aux ₂ ₃ = gen X₂₃-gen
  mX-aux ₃ ₄ = gen X₃₄-gen
  mX-aux ₄ ₅ = gen X₄₅-gen
  mX-aux ₅ ₆ = gen X₅₆-gen
  mX-aux ₆ ₇ = gen X₆₇-gen
  mX-aux _ _ = ε

  mX' : Nat -> Index -> Index -> Word Gen
  mX' 0 _ _ = ε
  mX' (succ n) l m with l =m? pred m
  ... | Nothing = mX-aux (pred m) m • mX' n l (pred m) • mX-aux (pred m) m
  ... | Just _ = mX-aux l m

  mX = mX' 100


  X₀₁ : Word Gen
  X₀₁ = gen X₀₁-gen

  X₁₂ : Word Gen
  X₁₂ = gen X₁₂-gen

  X₂₃ : Word Gen
  X₂₃ = gen X₂₃-gen

  X₃₄ : Word Gen
  X₃₄ = gen X₃₄-gen

  X₄₅ : Word Gen
  X₄₅ = gen X₄₅-gen

  X₅₆ : Word Gen
  X₅₆ = gen X₅₆-gen

  X₆₇ : Word Gen
  X₆₇ = gen X₆₇-gen

  X₀₂ : Word Gen
  X₀₂ = X₁₂ • X₀₁ • X₁₂

  X₀₃ : Word Gen
  X₀₃ = X₂₃ • X₀₂ • X₂₃

  X₀₄ : Word Gen
  X₀₄ = X₃₄ • X₀₃ • X₃₄

  X₀₅ : Word Gen
  X₀₅ = X₄₅ • X₀₄ • X₄₅

  X₀₆ : Word Gen
  X₀₆ = X₅₆ • X₀₅ • X₅₆

  X₀₇ : Word Gen
  X₀₇ = X₆₇ • X₀₆ • X₆₇

  X₁₃ : Word Gen
  X₁₃ = X₂₃ • X₁₂ • X₂₃

  X₁₄ : Word Gen
  X₁₄ = X₃₄ • X₁₃ • X₃₄

  X₁₅ : Word Gen
  X₁₅ = X₄₅ • X₁₄ • X₄₅

  X₁₆ : Word Gen
  X₁₆ = X₅₆ • X₁₅ • X₅₆

  X₁₇ : Word Gen
  X₁₇ = X₆₇ • X₁₆ • X₆₇

  X₂₄ : Word Gen
  X₂₄ = X₃₄ • X₂₃ • X₃₄

  X₂₅ : Word Gen
  X₂₅ = X₄₅ • X₂₄ • X₄₅

  X₂₆ : Word Gen
  X₂₆ = X₅₆ • X₂₅ • X₅₆

  X₂₇ : Word Gen
  X₂₇ = X₆₇ • X₂₆ • X₆₇

  X₃₅ : Word Gen
  X₃₅ = X₄₅ • X₃₄ • X₄₅

  X₃₆ : Word Gen
  X₃₆ = X₅₆ • X₃₅ • X₅₆

  X₃₇ : Word Gen
  X₃₇ = X₆₇ • X₃₆ • X₆₇

  X₄₆ : Word Gen
  X₄₆ = X₅₆ • X₄₅ • X₅₆

  X₄₇ : Word Gen
  X₄₇ = X₆₇ • X₄₆ • X₆₇

  X₅₇ : Word Gen
  X₅₇ = X₆₇ • X₅₆ • X₆₇

  mK' : Nat -> Index -> Index -> Word Gen
  mK' 0 _ _ = ε
  mK' (succ l) ₀ ₁ = gen K₀₁-gen
  mK' (succ l) ₀ n = mX (pred n) n • mK' l ₀ (pred n) • mX (pred n) n
  mK' (succ l) m n = mX (pred m) m • mK' l (pred m) n • mX (pred m) m

  mK = mK' 200


  i₀ : Word Gen
  i₀ = gen i₀-gen

  i₁ : Word Gen
  i₁ = X₀₁ • i₀ • X₀₁

  i₂ : Word Gen
  i₂ = X₁₂ • i₁ • X₁₂

  i₃ : Word Gen
  i₃ = X₂₃ • i₂ • X₂₃

  i₄ : Word Gen
  i₄ = X₃₄ • i₃ • X₃₄

  i₅ : Word Gen
  i₅ = X₄₅ • i₄ • X₄₅

  i₆ : Word Gen
  i₆ = X₅₆ • i₅ • X₅₆

  i₇ : Word Gen
  i₇ = X₆₇ • i₆ • X₆₇

  K₀₁ : Word Gen
  K₀₁ = gen K₀₁-gen

  K₁₂ : Word Gen
  K₁₂ = mK ₁ ₂

  K₂₃ : Word Gen
  K₂₃ = mK ₂ ₃

  K₃₄ : Word Gen
  K₃₄ = mK ₃ ₄

  K₄₅ : Word Gen
  K₄₅ = mK ₄ ₅

  K₅₆ : Word Gen
  K₅₆ = mK ₅ ₆

  K₆₇ : Word Gen
  K₆₇ = mK ₆ ₇

  K₀₂ : Word Gen
  K₀₂ = X₁₂ • K₀₁ • X₁₂

  K₀₃ : Word Gen
  K₀₃ = X₂₃ • K₀₂ • X₂₃

  K₀₄ : Word Gen
  K₀₄ = X₃₄ • K₀₃ • X₃₄

  K₀₅ : Word Gen
  K₀₅ = X₄₅ • K₀₄ • X₄₅

  K₀₆ : Word Gen
  K₀₆ = X₅₆ • K₀₅ • X₅₆

  K₀₇ : Word Gen
  K₀₇ = X₆₇ • K₀₆ • X₆₇

  K₁₃ : Word Gen
  K₁₃ = X₂₃ • K₁₂ • X₂₃

  K₁₄ : Word Gen
  K₁₄ = X₃₄ • K₁₃ • X₃₄

  K₁₅ : Word Gen
  K₁₅ = X₄₅ • K₁₄ • X₄₅

  K₁₆ : Word Gen
  K₁₆ = X₅₆ • K₁₅ • X₅₆

  K₁₇ : Word Gen
  K₁₇ = X₆₇ • K₁₆ • X₆₇

  K₂₄ : Word Gen
  K₂₄ = X₃₄ • K₂₃ • X₃₄

  K₂₅ : Word Gen
  K₂₅ = X₄₅ • K₂₄ • X₄₅

  K₂₆ : Word Gen
  K₂₆ = X₅₆ • K₂₅ • X₅₆

  K₂₇ : Word Gen
  K₂₇ = X₆₇ • K₂₆ • X₆₇

  K₃₅ : Word Gen
  K₃₅ = X₄₅ • K₃₄ • X₄₅

  K₃₆ : Word Gen
  K₃₆ = X₅₆ • K₃₅ • X₅₆

  K₃₇ : Word Gen
  K₃₇ = X₆₇ • K₃₆ • X₆₇

  K₄₆ : Word Gen
  K₄₆ = X₅₆ • K₄₅ • X₅₆

  K₄₇ : Word Gen
  K₄₇ = X₆₇ • K₄₆ • X₆₇

  K₅₇ : Word Gen
  K₅₇ = X₆₇ • K₅₆ • X₆₇


  -- The new set of relations.
  data Rel : Context Gen where
    [S1] : Rel (i₀ ^ 4 === ε)

    [S3a] : Rel (X₀₁ ^ 2 === ε)
    [S3b] : Rel (X₁₂ ^ 2 === ε)
    [S3c] : Rel (X₂₃ ^ 2 === ε)
    [S3d] : Rel (X₃₄ ^ 2 === ε)
    [S3e] : Rel (X₄₅ ^ 2 === ε)
    [S3f] : Rel (X₅₆ ^ 2 === ε)
    [S3g] : Rel (X₆₇ ^ 2 === ε)

    [S4a] : Rel (i₁ • i₀ === i₀ • i₁)
    [S4b] : Rel (i₂ • K₀₁ === K₀₁ • i₂)

    [S5a] : Rel (i₀ • X₁₂ === X₁₂ • i₀)
    [S5b] : Rel (i₀ • X₂₃ === X₂₃ • i₀)
    [S5c] : Rel (i₀ • X₃₄ === X₃₄ • i₀)
    [S5d] : Rel (i₀ • X₄₅ === X₄₅ • i₀)
    [S5e] : Rel (i₀ • X₅₆ === X₅₆ • i₀)
    [S5f] : Rel (i₀ • X₆₇ === X₆₇ • i₀)

    [S6] : Rel (K₂₃ • K₀₁ === K₀₁ • K₂₃)

    [S7a] : Rel (K₀₁ • X₂₃ === X₂₃ • K₀₁)
    [S7b] : Rel (K₀₁ • X₃₄ === X₃₄ • K₀₁)
    [S7c] : Rel (K₀₁ • X₄₅ === X₄₅ • K₀₁)
    [S7d] : Rel (K₀₁ • X₅₆ === X₅₆ • K₀₁)
    [S7e] : Rel (K₀₁ • X₆₇ === X₆₇ • K₀₁)

    [S8a] : Rel (X₂₃ • X₀₁ === X₀₁ • X₂₃)
    [S8b] : Rel (X₃₄ • X₀₁ === X₀₁ • X₃₄)
    [S8c] : Rel (X₄₅ • X₀₁ === X₀₁ • X₄₅)
    [S8d] : Rel (X₅₆ • X₀₁ === X₀₁ • X₅₆)
    [S8e] : Rel (X₆₇ • X₀₁ === X₀₁ • X₆₇)
    [S8f] : Rel (X₃₄ • X₁₂ === X₁₂ • X₃₄)
    [S8g] : Rel (X₄₅ • X₁₂ === X₁₂ • X₄₅)
    [S8h] : Rel (X₅₆ • X₁₂ === X₁₂ • X₅₆)
    [S8i] : Rel (X₆₇ • X₁₂ === X₁₂ • X₆₇)
    [S8j] : Rel (X₄₅ • X₂₃ === X₂₃ • X₄₅)
    [S8k] : Rel (X₅₆ • X₂₃ === X₂₃ • X₅₆)
    [S8l] : Rel (X₆₇ • X₂₃ === X₂₃ • X₆₇)
    [S8m] : Rel (X₅₆ • X₃₄ === X₃₄ • X₅₆)
    [S8n] : Rel (X₆₇ • X₃₄ === X₃₄ • X₆₇)
    [S8o] : Rel (X₆₇ • X₄₅ === X₄₅ • X₆₇)

    [S9a] : Rel (X₁₂ • X₀₁ • X₁₂ === X₀₁ • X₁₂ • X₀₁)
    [S9b] : Rel (X₂₃ • X₁₂ • X₂₃ === X₁₂ • X₂₃ • X₁₂)
    [S9c] : Rel (X₃₄ • X₂₃ • X₃₄ === X₂₃ • X₃₄ • X₂₃)
    [S9d] : Rel (X₄₅ • X₃₄ • X₄₅ === X₃₄ • X₄₅ • X₃₄)
    [S9e] : Rel (X₅₆ • X₄₅ • X₅₆ === X₄₅ • X₅₆ • X₄₅)
    [S9f] : Rel (X₆₇ • X₅₆ • X₆₇ === X₅₆ • X₆₇ • X₅₆)

    [S10] : Rel (K₀₁ • i₁ ^ 2 === X₀₁ • K₀₁)
    [S11] : Rel (i₁ • K₀₁ • i₁ • K₀₁ === K₀₁ • i₁ ^ 3)
    [S12] : Rel (K₀₁ • i₀ • i₁ === i₀ • i₁ • K₀₁)
    [S14] : Rel (K₀₁ ^ 2 • i₀ • i₁ === ε)
    [S15] : Rel (K₀₂ • K₁₃ • K₀₁ • K₂₃ === K₀₁ • K₂₃ • K₀₂ • K₁₃)


open TwoLevel-Simplified
open TwoLevel-Less

-- Translation from the new to the old generators.
f : TwoLevel-Simplified.Gen -> Word TwoLevel-Less.Gen
f i₀-gen = i ₀
f K₀₁-gen = K ₀₁
f X₀₁-gen = X ₀₁
f X₁₂-gen = X ₁₂
f X₂₃-gen = X ₂₃
f X₃₄-gen = X ₃₄
f X₄₅-gen = X ₄₅
f X₅₆-gen = X ₅₆
f X₆₇-gen = X ₆₇

-- ----------------------------------------------------------------------
-- * Statement of completeness and soundness of the new relations

soundness-property : Set
soundness-property = TwoLevel-Simplified.Rel is-sound-wrt TwoLevel-Less.Rel and f

completeness-property : Set
completeness-property = TwoLevel-Simplified.Rel is-complete-wrt TwoLevel-Less.Rel and f


-- We put the reverse translation here for convenience. The properties
-- stated do not depend on it.

-- Translation from the old to the new generators.
g : TwoLevel-Less.Gen -> Word TwoLevel-Simplified.Gen
g (i-gen ₀) = i₀
g (i-gen ₁) = i₁
g (i-gen ₂) = i₂
g (i-gen ₃) = i₃
g (i-gen ₄) = i₄
g (i-gen ₅) = i₅
g (i-gen ₆) = i₆
g (i-gen ₇) = i₇

g (X-gen {₀} {₁} jk) = X₀₁
g (X-gen {₀} {₂} jk) = X₀₂
g (X-gen {₀} {₃} jk) = X₀₃
g (X-gen {₀} {₄} jk) = X₀₄
g (X-gen {₀} {₅} jk) = X₀₅
g (X-gen {₀} {₆} jk) = X₀₆
g (X-gen {₀} {₇} jk) = X₀₇

g (X-gen {₁} {₂} jk) = X₁₂
g (X-gen {₁} {₃} jk) = X₁₃
g (X-gen {₁} {₄} jk) = X₁₄
g (X-gen {₁} {₅} jk) = X₁₅
g (X-gen {₁} {₆} jk) = X₁₆
g (X-gen {₁} {₇} jk) = X₁₇

g (X-gen {₂} {₃} jk) = X₂₃
g (X-gen {₂} {₄} jk) = X₂₄
g (X-gen {₂} {₅} jk) = X₂₅
g (X-gen {₂} {₆} jk) = X₂₆
g (X-gen {₂} {₇} jk) = X₂₇

g (X-gen {₃} {₄} jk) = X₃₄
g (X-gen {₃} {₅} jk) = X₃₅
g (X-gen {₃} {₆} jk) = X₃₆
g (X-gen {₃} {₇} jk) = X₃₇

g (X-gen {₄} {₅} jk) = X₄₅
g (X-gen {₄} {₆} jk) = X₄₆
g (X-gen {₄} {₇} jk) = X₄₇

g (X-gen {₅} {₆} jk) = X₅₆
g (X-gen {₅} {₇} jk) = X₅₇
g (X-gen {₅} {₅} ())
g (X-gen {₆} {₅} ())

g (X-gen {₆} {₆} ())
g (X-gen {₇} {₆} ())
g (X-gen {₆} {₇} _) = X₆₇
g (X-gen {₇} {₇} ())


g (K-gen {₀} {₁} jk) = K₀₁
g (K-gen {₀} {₂} jk) = K₀₂
g (K-gen {₀} {₃} jk) = K₀₃
g (K-gen {₀} {₄} jk) = K₀₄
g (K-gen {₀} {₅} jk) = K₀₅
g (K-gen {₀} {₆} jk) = K₀₆
g (K-gen {₀} {₇} jk) = K₀₇

g (K-gen {₁} {₂} jk) = K₁₂
g (K-gen {₁} {₃} jk) = K₁₃
g (K-gen {₁} {₄} jk) = K₁₄
g (K-gen {₁} {₅} jk) = K₁₅
g (K-gen {₁} {₆} jk) = K₁₆
g (K-gen {₁} {₇} jk) = K₁₇

g (K-gen {₂} {₃} jk) = K₂₃
g (K-gen {₂} {₄} jk) = K₂₄
g (K-gen {₂} {₅} jk) = K₂₅
g (K-gen {₂} {₆} jk) = K₂₆
g (K-gen {₂} {₇} jk) = K₂₇

g (K-gen {₃} {₄} jk) = K₃₄
g (K-gen {₃} {₅} jk) = K₃₅
g (K-gen {₃} {₆} jk) = K₃₆
g (K-gen {₃} {₇} jk) = K₃₇

g (K-gen {₄} {₅} jk) = K₄₅
g (K-gen {₄} {₆} jk) = K₄₆
g (K-gen {₄} {₇} jk) = K₄₇

g (K-gen {₅} {₆} jk) = K₅₆
g (K-gen {₅} {₇} jk) = K₅₇
g (K-gen {₅} {₅} ())
g (K-gen {₆} {₅} ())
g (K-gen {₆} {₆} ())
g (K-gen {₇} {₆} ())
g (K-gen {₆} {₇} jk) = K₆₇
g (K-gen {₇} {₇} ())

