open import Lib.Equality as Eq using (_==_ ; auto)
open import Lib.Nat
open import Lib.Maybe
open import Lib.Decidable
open import Lib.Boolean
open import Lib.Proposition
open import Lib.List
open import Lib.Word
open import Lib.Word-Lemmas
open import Lib.Word-Tactics

open Monoid-Equational
open Rewriting
open Associative
open InContext

open import Lib.MaybeEq
open import Lib.MaybeEq-Instances
open import CosetNF
open import Gate

open import Step7.Rel
open import Step7.MvI
open import Step7.Order
open import Step7.Basis-Change
open import Step7.PLemmas

module Step7.Comm where

  lemma-CX12-CX10=CX10-CX12 : Rel ⊢ CX12 • CX10 === CX10 • CX12
  lemma-CX12-CX10=CX10-CX12 = symm (lemma-CX10-CX12=CX12-CX10)
  lemma-CX20-CX10=CX10-CX20 : Rel ⊢ CX20 • CX10 === CX10 • CX20
  lemma-CX20-CX10=CX10-CX20 = symm (lemma-CX10-CX20=CX20-CX10)


  lemma-CX01-X1=X1-CX01 : Rel ⊢ CX01 • X1 === X1 • CX01
  lemma-CX01-X1=X1-CX01 = symm (B01.by-basis-change Swap01 Swap01 (axiom ax-X0-CX10=CX10-X0) 50 auto)

  lemma-CX01-X2=X2-CX01 : Rel ⊢ CX01 • X2 === X2 • CX01
  lemma-CX01-X2=X2-CX01 = symm (B02.by-basis-change (Swap12 • Swap01) (Swap01 • Swap12) (axiom ax-X0-CX12=CX12-X0) 50 auto)

  lemma-CX10-X0=X0-CX10 : Rel ⊢ CX10 • X0 === X0 • CX10
  lemma-CX10-X0=X0-CX10 = symm (axiom ax-X0-CX10=CX10-X0)

  lemma-CX10-X2=X2-CX10 : Rel ⊢ CX10 • X2 === X2 • CX10
  lemma-CX10-X2=X2-CX10 = B01.by-basis-change Swap01 Swap01 (lemma-CX01-X2=X2-CX01) 50 auto

  lemma-CX02-CX12=CX12-CX02 : Rel ⊢ CX02 • CX12 === CX12 • CX02
  lemma-CX02-CX12=CX12-CX02 = by-basis-change Swap12 Swap12 (axiom ax-CX01-CX21=CX21-CX01) 50 auto
  

  -- Commutativity.
  comm : (x y : Gate) -> Maybe (commutes Rel x y)

  comm CX01-gen CX21-gen = Just (axiom ax-CX01-CX21=CX21-CX01)
  comm CX01-gen CX02-gen = Just (axiom ax-CX01-CX02=CX02-CX01)
  comm CX01-gen X1-gen = Just (lemma-CX01-X1=X1-CX01)
  comm CX01-gen X2-gen = Just (lemma-CX01-X2=X2-CX01)
  comm CX01-gen S0-gen = Just (lemma-CX01-S0=S0-CX01)
  comm CX01-gen S2-gen = Just (lemma-CX01-S2=S2-CX01)
  comm CX01-gen CS02-gen = Just (lemma-CX01-CS02=CS02-CX01)
  comm CX01-gen iI-gen = Just (symm (axiom ax-iI-CX01=CX01-iI))
  comm CX01-gen CCX1-gen = Just (symm (B01.by-basis-change Swap01 Swap01 (axiom ax-CCX0-CX10=CX10-CCX0) 50 auto))
  
  comm CX10-gen CX12-gen = Just (lemma-CX10-CX12=CX12-CX10)
  comm CX10-gen CX20-gen = Just (lemma-CX10-CX20=CX20-CX10)
  comm CX10-gen X0-gen = Just (lemma-CX10-X0=X0-CX10)
  comm CX10-gen X2-gen = Just (lemma-CX10-X2=X2-CX10)
  comm CX10-gen S1-gen = Just (axiom ax-CX10-S1=S1-CX10)
  comm CX10-gen S2-gen = Just (axiom ax-CX10-S2=S2-CX10)
  comm CX10-gen CS12-gen = Just (axiom ax-CX10-CS12=CS12-CX10)
  comm CX10-gen iI-gen = Just (symm (lemma-iI-CX10=CX10-iI))
  comm CX10-gen CCX0-gen = Just (symm (axiom ax-CCX0-CX10=CX10-CCX0))
  
  comm CX12-gen CX10-gen = Just (symm (lemma-CX10-CX12=CX12-CX10))
  comm CX12-gen CX02-gen = Just (lemma-CX12-CX02=CX02-CX12)
  comm CX12-gen X0-gen = Just (symm (axiom ax-X0-CX12=CX12-X0))
  comm CX12-gen X2-gen = Just (B02.by-basis-change (Swap12 • Swap01 • Swap12) (Swap12 • Swap01 • Swap12) (lemma-CX10-X0=X0-CX10) 50 auto)
  comm CX12-gen S0-gen = Just (B01.by-basis-change Swap01 Swap01 (lemma-CX02-S1=S1-CX02) 50 auto)
  comm CX12-gen S1-gen = Just (B02.by-basis-change (Swap01 • Swap12) (Swap12 • Swap01) (lemma-CX01-S0=S0-CX01) 100 auto)
  comm CX12-gen CS01-gen = Just lemma-CX12-CS01=CS01-CX12
  comm CX12-gen iI-gen = Just (symm (lemma-iI-CX12=CX12-iI))
  comm CX12-gen K0-gen = Just (axiom ax-CX12-K0=K0-CX12)
  comm CX12-gen CCX2-gen = Just (symm (lemma-CCX2-CX12=CX12-CCX2))
  
  comm CX21-gen CX01-gen = Just (symm (axiom ax-CX01-CX21=CX21-CX01))
  comm CX21-gen CX20-gen = Just (lemma-CX21-CX20=CX20-CX21)
  comm CX21-gen X0-gen = Just (symm (lemma-X0-CX21=CX21-X0))
  comm CX21-gen X1-gen = Just (symm (lemma-X1-CX21=CX21-X1))
  comm CX21-gen S0-gen = Just (B02.by-basis-change (Swap12 • Swap01) (Swap01 • Swap12) (lemma-CX02-S1=S1-CX02) 50 auto)
  comm CX21-gen S2-gen = Just (B02.by-basis-change (Swap01 • Swap12) (Swap12 • Swap01) (axiom ax-CX10-S1=S1-CX10) 50 auto)
  comm CX21-gen CS02-gen = Just (B01.by-basis-change Swap01 Swap01 (lemma-CX20-CS12=CS12-CX20) 50 auto)
  comm CX21-gen iI-gen = Just (symm (lemma-iI-CX21=CX21-iI))
  comm CX21-gen K0-gen = Just (lemma-CX21-K0=K0-CX21)
  comm CX21-gen CCX1-gen = Just (symm (B01.by-basis-change Swap01 Swap01 (lemma-CCX0-CX20=CX20-CCX0) 50 auto))
  
  comm CX02-gen CX01-gen = Just (symm (axiom ax-CX01-CX02=CX02-CX01))
  comm CX02-gen CX12-gen = Just (lemma-CX02-CX12=CX12-CX02)
  comm CX02-gen X1-gen = Just (by-basis-change Swap12 Swap12 (lemma-CX01-X2=X2-CX01) 50 auto)
  comm CX02-gen X2-gen = Just (by-basis-change Swap12 Swap12 (lemma-CX01-X1=X1-CX01) 50 auto)
  comm CX02-gen S0-gen = Just (by-basis-change Swap12 Swap12 (lemma-CX01-S0=S0-CX01) 50 auto)
  comm CX02-gen S1-gen = Just (lemma-CX02-S1=S1-CX02)
  comm CX02-gen CS01-gen = Just (lemma-CX02-CS01=CS01-CX02)
  comm CX02-gen iI-gen = Just (symm (lemma-iI-CX02=CX02-iI))
  comm CX02-gen CCX2-gen = Just (symm (lemma-CCX2-CX02=CX02-CCX2))
  
  comm CX20-gen CX10-gen = Just (symm (lemma-CX10-CX20=CX20-CX10))
  comm CX20-gen CX21-gen = Just (symm (lemma-CX21-CX20=CX20-CX21))
  comm CX20-gen X0-gen = Just (symm (lemma-X0-CX20=CX20-X0))
  comm CX20-gen X1-gen = Just (symm (lemma-X1-CX20=CX20-X1))
  comm CX20-gen S1-gen = Just (lemma-CX20-S1=S1-CX20)
  comm CX20-gen S2-gen = Just (by-basis-change Swap12 Swap12 (axiom ax-CX10-S1=S1-CX10) 50 auto)
  comm CX20-gen iI-gen = Just (symm (lemma-iI-CX20=CX20-iI))
  comm CX20-gen CCX0-gen = Just (symm (lemma-CCX0-CX20=CX20-CCX0))

  comm X0-gen CCX0-gen = Just (axiom ax-X0-CCX0=CCX0-X0)
  comm X0-gen CX10-gen = Just (axiom ax-X0-CX10=CX10-X0)
  comm X0-gen CX12-gen = Just (axiom ax-X0-CX12=CX12-X0)
  comm X0-gen CX21-gen = Just (lemma-X0-CX21=CX21-X0)
  comm X0-gen CX20-gen = Just (lemma-X0-CX20=CX20-X0)
  comm X0-gen X1-gen = Just (symm (axiom ax-X1-X0=X0-X1))
  comm X0-gen X2-gen = Just (symm (lemma-X2-X0=X0-X2))
  comm X0-gen Swap12-gen = Just (symm (axiom ax-Swap12-X0=X0-Swap12))
  comm X0-gen S1-gen = Just (axiom ax-X0-S1=S1-X0)
  comm X0-gen S2-gen = Just (by-basis-change Swap12 Swap12 (axiom ax-X0-S1=S1-X0) 50 auto)
  comm X0-gen CS12-gen = Just (axiom ax-X0-CS12=CS12-X0)
  comm X0-gen iI-gen = Just (symm (axiom ax-iI-X0=X0-iI))
  
  comm X1-gen CCX1-gen = Just (lemma-X1-CCX1=CCX1-X1)
  comm X1-gen CX01-gen = Just (lemma-X1-CX01=CX01-X1)
  comm X1-gen CX21-gen = Just (lemma-X1-CX21=CX21-X1)
  comm X1-gen CX02-gen = Just (symm (by-basis-change Swap12 Swap12 (lemma-CX01-X2=X2-CX01) 50 auto))
  comm X1-gen CX20-gen = Just ( lemma-X1-CX20=CX20-X1)
  comm X1-gen X0-gen = Just (axiom ax-X1-X0=X0-X1)
  comm X1-gen X2-gen = Just (symm (lemma-X2-X1=X1-X2))
  comm X1-gen S0-gen = Just (lemma-X1-S0=S0-X1)
  comm X1-gen S2-gen = Just (lemma-X1-S2=S2-X1)
  comm X1-gen CS02-gen = Just (lemma-X1-CS02=CS02-X1)
  comm X1-gen iI-gen = Just (symm (lemma-iI-X1=X1-iI))
  comm X1-gen K0-gen = Just (axiom ax-X1-K0=K0-X1)
  
  comm X2-gen CCX2-gen = Just (lemma-X2-CCX2=CCX2-X2)
  comm X2-gen CX01-gen = Just (lemma-X2-CX01=CX01-X2)
  comm X2-gen CX10-gen = Just (lemma-X2-CX10=CX10-X2)
  comm X2-gen CX12-gen = Just (symm (B02.by-basis-change (Swap12 • Swap01 • Swap12) (Swap12 • Swap01 • Swap12) (lemma-CX10-X0=X0-CX10) 50 auto))
  comm X2-gen CX02-gen = Just (symm (by-basis-change Swap12 Swap12 (lemma-CX01-X1=X1-CX01) 50 auto))
  comm X2-gen X0-gen = Just (lemma-X2-X0=X0-X2)
  comm X2-gen X1-gen = Just (lemma-X2-X1=X1-X2)
  comm X2-gen Swap01-gen = Just (symm (axiom ax-Swap01-X2=X2-Swap01))
  comm X2-gen S0-gen = Just (by-basis-change Swap12 Swap12 (lemma-X1-S0=S0-X1) 50 auto)
  comm X2-gen S1-gen = Just (B02.by-basis-change (Swap01 • Swap12) (Swap12 • Swap01) (lemma-X1-S0=S0-X1) 50 auto)
  comm X2-gen CS01-gen = Just (by-basis-change Swap12 Swap12 (lemma-X1-CS02=CS02-X1) 50 auto)
  comm X2-gen iI-gen = Just (symm (lemma-iI-X2=X2-iI))
  comm X2-gen K0-gen = Just (lemma-X2-K0=K0-X2)
  
  comm Swap01-gen CCX2-gen = Just (axiom ax-Swap01-CCX2=CCX2-Swap01)
  comm Swap01-gen X2-gen = Just (axiom ax-Swap01-X2=X2-Swap01)
  comm Swap01-gen S2-gen = Just (axiom ax-Swap01-S2=S2-Swap01)
  comm Swap01-gen CS01-gen = Just (axiom ax-Swap01-CS01=CS01-Swap01)
  comm Swap01-gen CCZ-gen = Just (axiom ax-Swap01-CCZ=CCZ-Swap01)
  comm Swap01-gen iI-gen = Just (symm (axiom ax-iI-Swap01=Swap01-iI))
  
  comm Swap12-gen CCX0-gen = Just (axiom ax-Swap12-CCX0=CCX0-Swap12)
  comm Swap12-gen X0-gen = Just (axiom ax-Swap12-X0=X0-Swap12)
  comm Swap12-gen S0-gen = Just (axiom ax-Swap12-S0=S0-Swap12)
  comm Swap12-gen CS12-gen = Just (axiom ax-Swap12-CS12=CS12-Swap12)
  comm Swap12-gen CCZ-gen = Just (axiom ax-Swap12-CCZ=CCZ-Swap12)
  comm Swap12-gen iI-gen = Just (symm (axiom ax-iI-Swap12=Swap12-iI))
  comm Swap12-gen K0-gen = Just (axiom ax-Swap12-K0=K0-Swap12)
  
  comm S0-gen CCX1-gen = Just (symm (lemma-CCX1-S0=S0-CCX1))
  comm S0-gen CCX2-gen = Just (symm (by-basis-change Swap12 Swap12 (lemma-CCX1-S0=S0-CCX1) 50 auto))
  comm S0-gen CX01-gen = Just (symm (lemma-CX01-S0=S0-CX01))
  comm S0-gen CX12-gen = Just (symm (B01.by-basis-change Swap01 Swap01 (lemma-CX02-S1=S1-CX02) 50 auto))
  comm S0-gen CX21-gen = Just (symm (B02.by-basis-change (Swap12 • Swap01) (Swap01 • Swap12) (lemma-CX02-S1=S1-CX02) 50 auto))
  comm S0-gen CX02-gen = Just (symm (by-basis-change Swap12 Swap12 (lemma-CX01-S0=S0-CX01) 50 auto))
  comm S0-gen Swap12-gen = Just (symm (axiom ax-Swap12-S0=S0-Swap12))
  comm S0-gen X1-gen = Just (symm (lemma-X1-S0=S0-X1))
  comm S0-gen X2-gen = Just (symm (by-basis-change Swap12 Swap12 (lemma-X1-S0=S0-X1) 50 auto))
  comm S0-gen S1-gen = Just (symm (axiom ax-S1-S0=S0-S1))
  comm S0-gen S2-gen = Just (symm (by-basis-change Swap12 Swap12 (axiom ax-S1-S0=S0-S1) 50 auto))
  comm S0-gen CS01-gen = Just (symm (axiom ax-CS01-S0=S0-CS01))
  comm S0-gen CS12-gen = Just (symm (lemma-CS12-S0=S0-CS12))
  comm S0-gen CS02-gen = Just (symm (by-basis-change Swap12 Swap12 (axiom ax-CS01-S0=S0-CS01) 50 auto))
  comm S0-gen CCZ-gen = Just (symm (axiom ax-CCZ-S0=S0-CCZ))
  comm S0-gen iI-gen = Just (symm (axiom ax-iI-S0=S0-iI))
  
  comm S1-gen CCX0-gen = Just (symm (axiom ax-CCX0-S1=S1-CCX0))
  comm S1-gen CCX2-gen = Just (symm (lemma-CCX2-S1=S1-CCX2))
  comm S1-gen CX10-gen = Just (symm (axiom ax-CX10-S1=S1-CX10))
  comm S1-gen CX12-gen = Just (symm (B02.by-basis-change (Swap01 • Swap12) (Swap12 • Swap01) (lemma-CX01-S0=S0-CX01) 100 auto))
  comm S1-gen CX02-gen = Just (symm (lemma-CX02-S1=S1-CX02))
  comm S1-gen CX20-gen = Just (symm (lemma-CX20-S1=S1-CX20))
  comm S1-gen X0-gen = Just (symm (axiom ax-X0-S1=S1-X0))
  comm S1-gen X2-gen = Just (symm (B02.by-basis-change (Swap01 • Swap12) (Swap12 • Swap01) (lemma-X1-S0=S0-X1) 50 auto))
  comm S1-gen S0-gen = Just (axiom ax-S1-S0=S0-S1)
  comm S1-gen S2-gen = Just (symm (lemma-S2-S1=S1-S2))
  comm S1-gen CS01-gen = Just (symm (lemma-CS01-S1=S1-CS01))
  comm S1-gen CS12-gen = Just (symm (lemma-CS12-S1=S1-CS12))
  comm S1-gen CS02-gen = Just (symm (lemma-CS02-S1=S1-CS02))
  comm S1-gen CCZ-gen = Just (symm (lemma-CCZ-S1=S1-CCZ))
  comm S1-gen iI-gen = Just (symm (lemma-iI-S1=S1-iI))
  comm S1-gen K0-gen = Just (axiom ax-S1-K0=K0-S1)
  
  comm S2-gen CCX0-gen = Just (symm (lemma-CCX0-S2=S2-CCX0))
  comm S2-gen CCX1-gen = Just (symm (lemma-CCX1-S2=S2-CCX1))
  comm S2-gen CX01-gen = Just (symm (lemma-CX01-S2=S2-CX01))
  comm S2-gen CX10-gen = Just (symm (axiom ax-CX10-S2=S2-CX10))
  comm S2-gen CX21-gen = Just (symm (B02.by-basis-change (Swap01 • Swap12) (Swap12 • Swap01) (axiom ax-CX10-S1=S1-CX10) 50 auto))
  comm S2-gen CX20-gen = Just (symm (by-basis-change Swap12 Swap12 (axiom ax-CX10-S1=S1-CX10) 50 auto))
  comm S2-gen X0-gen = Just (symm (by-basis-change Swap12 Swap12 (axiom ax-X0-S1=S1-X0) 50 auto))
  comm S2-gen X1-gen = Just (symm (lemma-X1-S2=S2-X1))
  comm S2-gen Swap01-gen = Just (symm (axiom ax-Swap01-S2=S2-Swap01))
  comm S2-gen S0-gen = Just (by-basis-change Swap12 Swap12 (axiom ax-S1-S0=S0-S1) 50 auto)
  comm S2-gen S1-gen = Just (lemma-S2-S1=S1-S2)
  comm S2-gen CS01-gen = Just (symm (axiom ax-CS01-S2=S2-CS01))
  comm S2-gen CS12-gen = Just (symm (lemma-CS12-S2=S2-CS12))
  comm S2-gen CS02-gen = Just (symm (by-basis-change Swap12 Swap12 (lemma-CS01-S1=S1-CS01) 50 auto))
  comm S2-gen CCZ-gen = Just (symm (by-basis-change Swap12 Swap12 (lemma-CCZ-S1=S1-CCZ) 50 auto))
  comm S2-gen iI-gen = Just (symm (lemma-iI-S2=S2-iI))
  comm S2-gen K0-gen = Just (lemma-S2-K0=K0-S2)
  
  comm CS01-gen CCX2-gen = Just (symm (lemma-CCX2-CS01=CS01-CCX2))
  comm CS01-gen CX12-gen = Just (symm (B01.by-basis-change Swap01 Swap01 (lemma-CX02-CS01=CS01-CX02) 50 auto))
  comm CS01-gen CX02-gen = Just (symm (lemma-CX02-CS01=CS01-CX02))
  comm CS01-gen X2-gen = Just (symm (by-basis-change Swap12 Swap12 (lemma-X1-CS02=CS02-X1) 50 auto))
  comm CS01-gen Swap01-gen = Just (symm (axiom ax-Swap01-CS01=CS01-Swap01))
  comm CS01-gen S0-gen = Just (axiom ax-CS01-S0=S0-CS01)
  comm CS01-gen S1-gen = Just (lemma-CS01-S1=S1-CS01)
  comm CS01-gen S2-gen = Just (axiom ax-CS01-S2=S2-CS01)
  comm CS01-gen CS12-gen = Just (symm (axiom ax-CS12-CS01=CS01-CS12))
  comm CS01-gen CS02-gen = Just (symm (lemma-CS02-CS01=CS01-CS02))
  comm CS01-gen CCZ-gen = Just (symm (axiom ax-CCZ-CS01=CS01-CCZ))
  comm CS01-gen iI-gen = Just (symm (axiom ax-iI-CS01=CS01-iI))
  
  comm CS12-gen CCX0-gen = Just (symm (axiom ax-CCX0-CS12=CS12-CCX0))
  comm CS12-gen CX10-gen = Just (symm (axiom ax-CX10-CS12=CS12-CX10))
  comm CS12-gen CX20-gen = Just (symm (lemma-CX20-CS12=CS12-CX20))
  comm CS12-gen X0-gen = Just (symm (axiom ax-X0-CS12=CS12-X0))
  comm CS12-gen Swap12-gen = Just (symm (axiom ax-Swap12-CS12=CS12-Swap12))
  comm CS12-gen S0-gen = Just (lemma-CS12-S0=S0-CS12)
  comm CS12-gen S1-gen = Just (lemma-CS12-S1=S1-CS12)
  comm CS12-gen S2-gen = Just (lemma-CS12-S2=S2-CS12)
  comm CS12-gen CS01-gen = Just (axiom ax-CS12-CS01=CS01-CS12)
  comm CS12-gen CS02-gen = Just (lemma-CS12-CS02=CS02-CS12)
  comm CS12-gen CCZ-gen = Just (symm (lemma-CCZ-CS12=CS12-CCZ))
  comm CS12-gen iI-gen = Just (symm (mvID.general-rewrite 40 auto))
  comm CS12-gen K0-gen = Just (axiom ax-CS12-K0=K0-CS12)
  
  comm CS02-gen CCX1-gen = Just (symm (lemma-CCX1-CS02=CS02-CCX1))
  comm CS02-gen CX01-gen = Just (symm (lemma-CX01-CS02=CS02-CX01))
  comm CS02-gen CX21-gen = Just (symm (B01.by-basis-change Swap01 Swap01 (lemma-CX20-CS12=CS12-CX20) 50 auto))
  comm CS02-gen X1-gen = Just (symm (lemma-X1-CS02=CS02-X1))
  comm CS02-gen S0-gen = Just (by-basis-change Swap12 Swap12 (axiom ax-CS01-S0=S0-CS01) 50 auto)
  comm CS02-gen S1-gen = Just (lemma-CS02-S1=S1-CS02)
  comm CS02-gen S2-gen = Just (by-basis-change Swap12 Swap12 (lemma-CS01-S1=S1-CS01) 50 auto)
  comm CS02-gen CS01-gen = Just (lemma-CS02-CS01=CS01-CS02)
  comm CS02-gen CS12-gen = Just (symm (lemma-CS12-CS02=CS02-CS12))
  comm CS02-gen CCZ-gen = Just (symm (by-basis-change Swap12 Swap12 (axiom ax-CCZ-CS01=CS01-CCZ) 50 auto))
  comm CS02-gen iI-gen = Just (symm (lemma-iI-CS02=CS02-iI))
  
  comm CCZ-gen Swap01-gen = Just (symm (axiom ax-Swap01-CCZ=CCZ-Swap01))
  comm CCZ-gen Swap12-gen = Just (symm (axiom ax-Swap12-CCZ=CCZ-Swap12))
  comm CCZ-gen S0-gen = Just (axiom ax-CCZ-S0=S0-CCZ)
  comm CCZ-gen S1-gen = Just (lemma-CCZ-S1=S1-CCZ)
  comm CCZ-gen S2-gen = Just (by-basis-change Swap12 Swap12 (lemma-CCZ-S1=S1-CCZ) 50 auto)
  comm CCZ-gen CS01-gen = Just (axiom ax-CCZ-CS01=CS01-CCZ)
  comm CCZ-gen CS12-gen = Just (lemma-CCZ-CS12=CS12-CCZ)
  comm CCZ-gen CS02-gen = Just (by-basis-change Swap12 Swap12 (axiom ax-CCZ-CS01=CS01-CCZ) 50 auto)
  comm CCZ-gen iI-gen = Just (symm (axiom ax-iI-CCZ=CCZ-iI))
  
  comm iI-gen CCX0-gen = Just (axiom ax-iI-CCX0=CCX0-iI)
  comm iI-gen CCX1-gen = Just (lemma-iI-CCX1=CCX1-iI)
  comm iI-gen CCX2-gen = Just (lemma-iI-CCX2=CCX2-iI)
  comm iI-gen CX01-gen = Just (axiom ax-iI-CX01=CX01-iI)
  comm iI-gen CX10-gen = Just (lemma-iI-CX10=CX10-iI)
  comm iI-gen CX12-gen = Just (lemma-iI-CX12=CX12-iI)
  comm iI-gen CX21-gen = Just (lemma-iI-CX21=CX21-iI)
  comm iI-gen CX02-gen = Just (lemma-iI-CX02=CX02-iI)
  comm iI-gen CX20-gen = Just (lemma-iI-CX20=CX20-iI)
  comm iI-gen X0-gen = Just (axiom ax-iI-X0=X0-iI)
  comm iI-gen X1-gen = Just (lemma-iI-X1=X1-iI)
  comm iI-gen X2-gen = Just (lemma-iI-X2=X2-iI)
  comm iI-gen Swap01-gen = Just (axiom ax-iI-Swap01=Swap01-iI)
  comm iI-gen Swap12-gen = Just (axiom ax-iI-Swap12=Swap12-iI)
  comm iI-gen S0-gen = Just (axiom ax-iI-S0=S0-iI)
  comm iI-gen S1-gen = Just (lemma-iI-S1=S1-iI)
  comm iI-gen S2-gen = Just (lemma-iI-S2=S2-iI)
  comm iI-gen CS01-gen = Just (axiom ax-iI-CS01=CS01-iI)
  comm iI-gen CS12-gen = Just (mvID.general-rewrite 20 auto)
  comm iI-gen CS02-gen = Just (lemma-iI-CS02=CS02-iI)
  comm iI-gen CCZ-gen = Just (axiom ax-iI-CCZ=CCZ-iI)
  comm iI-gen K0-gen = Just (axiom ax-iI-K0=K0-iI)
  
  comm K0-gen CX12-gen = Just (symm (axiom ax-CX12-K0=K0-CX12))
  comm K0-gen CX21-gen = Just (symm (lemma-CX21-K0=K0-CX21))
  comm K0-gen X1-gen = Just (symm (axiom ax-X1-K0=K0-X1))
  comm K0-gen X2-gen = Just (symm (lemma-X2-K0=K0-X2))
  comm K0-gen Swap12-gen = Just (symm (axiom ax-Swap12-K0=K0-Swap12))
  comm K0-gen S1-gen = Just (symm (axiom ax-S1-K0=K0-S1))
  comm K0-gen S2-gen = Just (symm (lemma-S2-K0=K0-S2))
  comm K0-gen CS12-gen = Just (symm (axiom ax-CS12-K0=K0-CS12))
  comm K0-gen iI-gen = Just (symm (axiom ax-iI-K0=K0-iI))
  
  comm CCX0-gen CX10-gen = Just (axiom ax-CCX0-CX10=CX10-CCX0)
  comm CCX0-gen CX20-gen = Just (lemma-CCX0-CX20=CX20-CCX0)
  comm CCX0-gen X0-gen = Just (symm (axiom ax-X0-CCX0=CCX0-X0))
  comm CCX0-gen Swap12-gen = Just (symm (axiom ax-Swap12-CCX0=CCX0-Swap12))
  comm CCX0-gen S1-gen = Just (axiom ax-CCX0-S1=S1-CCX0)
  comm CCX0-gen S2-gen = Just (lemma-CCX0-S2=S2-CCX0)
  comm CCX0-gen CS12-gen = Just (axiom ax-CCX0-CS12=CS12-CCX0)
  comm CCX0-gen iI-gen = Just (symm (axiom ax-iI-CCX0=CCX0-iI))
  
  comm CCX1-gen CX01-gen = Just (B01.by-basis-change Swap01 Swap01 (axiom ax-CCX0-CX10=CX10-CCX0) 50 auto)
  comm CCX1-gen CX21-gen = Just (B01.by-basis-change Swap01 Swap01 (lemma-CCX0-CX20=CX20-CCX0) 50 auto)
  comm CCX1-gen X1-gen = Just (symm (lemma-X1-CCX1=CCX1-X1))
  comm CCX1-gen S0-gen = Just (lemma-CCX1-S0=S0-CCX1)
  comm CCX1-gen S2-gen = Just (lemma-CCX1-S2=S2-CCX1)
  comm CCX1-gen CS02-gen = Just (lemma-CCX1-CS02=CS02-CCX1)
  comm CCX1-gen iI-gen = Just (symm (lemma-iI-CCX1=CCX1-iI))
  
  comm CCX2-gen CX12-gen = Just (lemma-CCX2-CX12=CX12-CCX2)
  comm CCX2-gen CX02-gen = Just (lemma-CCX2-CX02=CX02-CCX2)
  comm CCX2-gen X2-gen = Just (symm (lemma-X2-CCX2=CCX2-X2))
  comm CCX2-gen Swap01-gen = Just (symm (axiom ax-Swap01-CCX2=CCX2-Swap01))
  comm CCX2-gen S0-gen = Just (by-basis-change Swap12 Swap12 (lemma-CCX1-S0=S0-CCX1) 50 auto)
  comm CCX2-gen S1-gen = Just (lemma-CCX2-S1=S1-CCX2)
  comm CCX2-gen CS01-gen = Just (lemma-CCX2-CS01=CS01-CCX2)
  comm CCX2-gen iI-gen = Just (symm (lemma-iI-CCX2=CCX2-iI))
  comm x y = Nothing

  -- We number the generators for the purpose of ordering them.
  ord : Gate -> Nat
  ord CCX0-gen = 28
  ord CCX1-gen = 27
  ord CCX2-gen = 26
  ord CX01-gen = 25
  ord CX10-gen = 24
  ord CX12-gen = 23
  ord CX21-gen = 22
  ord CX02-gen = 21
  ord CX20-gen = 20
  ord X0-gen = 19
  ord X1-gen = 18
  ord X2-gen = 17
  ord Swap01-gen = 16
  ord Swap12-gen = 15
  ord S0-gen = 14
  ord S1-gen = 13
  ord S2-gen = 12
  ord CS01-gen = 11
  ord CS12-gen = 10
  ord CS02-gen = 9
  ord CCZ-gen = 8
  ord iI-gen = 7
  ord CCK'-gen = 6
  ord CK10-gen = 5
  ord CK20-gen = 4
  ord K0-gen = 3
  ord K1-gen = 2
  ord K2-gen = 1
  
  -- Ordering of generators.
  less : Gate -> Gate -> Boolean
  less x y with ord x <=? ord y
  less x y | yes _ = true
  less x y | no _ = false
  
  open Commuting Gate Rel comm less public

  nf-comm : ListNF Rel
  nf-comm = record { listnf = comm-canonical ; lemma-listnf = lemma-comm-canonical }
